/*  Automat komrkowy Chaos Engine         */
/*  Wersja PocketPC GAPI                   */
/*  Piotr Gawrysiak, 2002                  */
/*  pgawrysiak@supermedia.pl               */
/*  http://bolek.ii.pw.edu.pl/~gawrysia    */

/*
Copyright (C) 2002 Piotr Gawrysiak

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//Makra

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include <windows.h>
#include <gx.h>
#include <stdio.h>
#include <stdarg.h>

#define TITLE TEXT("Chaos engine")

//Zmienne globalne

GXDisplayProperties gx_displayprop;

unsigned short Palette[256];
unsigned short Plytka1[240*240];
unsigned short Plytka2[240*240];
unsigned short backbuffer[240*320]; //To jest bufor obrazu, patrz funkcja UpdateFrame

//Funkcje


//Oblicza warto komrki pamici ekranu na podstawie skadowych koloru
unsigned short ComputeColor(int r, int g, int b)
{

	unsigned short PixelCol;
	if (gx_displayprop.ffFormat & kfDirect565)
	{
		PixelCol = (unsigned short) (r &0x00ff ) << 11 | (g & 0x00ff) << 6 | (b & 0x00ff);
	}
	else
	{
		PixelCol = (unsigned short) (r & 0xff) << 10 | (g & 0xff) << 5 | (b & 0xff);
	}

	return PixelCol;
}//of ComputeColor


//Generuje kolejny krok symulacji i odwiea zawarto ekranu
static void UpdateFrame()
{
	unsigned short x, y;
	unsigned short suma;
	long helper;

	//Uywajc wartoci komrek z tablicy Plytka1 generujemy wartoci dla nastpnego kroku symulacji
	//zapisujc je w tablicy Plytka2. Po zakoczeniu oblicze musimy przepisa zawarto Plytka2 do Plytka1

	//Zauwamy, e obliczenia przeprowadzamy dla komrek nie nalecych do brzegu pytki. Dziki temu krawd
	//ma zawsze warto 0, a ponadto nie mamy problemw z badaniem otoczenia komrki
	for (y=1;y<240;y++)
	{
		for(x=1;x<240;x++)
		{
			suma=0;

			//Najpierw sumujemy wartoci wszystkich ssiadujcych komrek i warto analizowanej komrki
			suma+=Plytka1[y*240+x];
			suma+=Plytka1[(y+1)*240+x];
			suma+=Plytka1[(y+1)*240+x+1];
			suma+=Plytka1[(y+1)*240+x-1];

			suma+=Plytka1[y*240+x+1];
			suma+=Plytka1[y*240+x-1];

			suma+=Plytka1[(y-1)*240+x];
			suma+=Plytka1[(y-1)*240+x+1];
			suma+=Plytka1[(y-1)*240+x-1];

			//I troszk si bawimy...
			//Jeli zastpimy kolejne cztery linie zwykym dzieleniem przez 9
			//to oblicza bdziemy zwyk redni arytmetyczn co daoby klasyczn symulacj stygnicia pytki :-) 
			helper=suma%9;
			suma = suma/9;
			if (suma>helper) suma++;
			suma = suma & 0x00FF;

			if (suma>255) suma=255;

			Plytka2[y*240+x]=suma;
		}

	}

	//Kopiujemy zawrto tablicy Plytka2 do Plytka1
	memcpy(Plytka1, Plytka2, sizeof(unsigned short)*240*240);

	//Pobieramy adres pocztku pamici obrazu
	unsigned short * buffer  = (unsigned short *) GXBeginDraw();
	
	//Tu w zasadzie powinnimy wywietli jaki komunikat o bdzie...
	if (buffer == NULL) return;

	//Kopiujemy zawarto tablicy Plytka2 do pamici ekranu
	//Nie moemy bezporednio przenie zawartoci tablicy Plytka2 do pamici ekranu np. funkcj memcpy
	//bo w wikszo PDA pami ta nie jest zorganizowana liniowo. Dlatego te musimy uywa wartoci
	//cbxPitch i cbyPitch. 
	//Ehhh, gdyby tak PDA miay sprztowy blitter...:-)
	//Przy okazji dokonujemy przypisania kolorw odpowiednim wartociom komrek, poprzez tablic Palette
	
	for(y=0;y<240;y++)
		for(x=0;x<240;x++)
			buffer[(x * gx_displayprop.cbxPitch>>1) + (y * gx_displayprop.cbyPitch>>1)]=Palette[Plytka1[y*240+x]];
	
	
	//Jeszcze jedna uwaga - tak naprawde przydaby si tu double buffering, bo powysza operacja
	//jest do powolna (a moe by jeszcze wolniejsza, jeli stosowane przez nas operacje graficzne bd bardziej skomplikowane. 
	//Aby to jednak uczyni naleaoby obliczy prawdziw warto pocztku pamici
	//obrazu - zmienna buffer zawiera bowiem jedynie adres pierwszego piksela, za wartoci cbxPitch i/lub
	//cbyPitch mog by - jak to si dzieje np. w iPaq'u 3630 - ujemne.
	//Mogoby to wyglda np. tak (cho jest oczywicie kod mocno niezoptymalizowany :-):
	/*
	//Zakladamy, ze jedynie cbyPitch moe by mniejsze od 0, co jest by moe nieco zbyt optymistyczne...

	//Obliczenie pocztku pamici ekranu - to te powinno robi si tylko raz...
	unsigned short * memstart = backbuffer;
	if (gx_displayprop.cbyPitch<0)
	{
		buffer = buffer + ((gx_displayprop.cyHeight-1)*(gx_displayprop.cbyPitch>>1));
		memstart = (unsigned short *)backbuffer + ((gx_displayprop.cyHeight-1)*-1*(gx_displayprop.cbyPitch>>1));
	}

	//Kopiowanie do bufora obrazu
	for(y=0;y<240;y++)
		for(x=0;x<240;x++)
			memstart[(x * gx_displayprop.cbxPitch>>1) + (y * gx_displayprop.cbyPitch>>1)]=Palette[Plytka1[y*240+x]];
			
	//Kopiowanie do pamici ekranu
	memcpy(buffer, backbuffer, 240*320*sizeof(unsigned short));
	*/
	
	GXEndDraw();

}//of UpdateFrame

long FAR PASCAL WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{

    switch (message)
    {

	case WM_LBUTTONDOWN:
             //Koczymy dziaanie programu po pukniciu rysikiem
             PostMessage(hWnd, WM_CLOSE, 0, 0);
             return 0L;

        case WM_DESTROY:
            PostQuitMessage(0);
            return 0L;

        case WM_SETCURSOR:
            //Na wszelki wypadek wyczamy ewentualny kursor myszki... :-)
            SetCursor(NULL);
            return TRUE;

    }
    return DefWindowProc(hWnd, message, wParam, lParam);
}//of WindowProc


//Inicjalizacja
static HRESULT InitApp(HINSTANCE hInstance, int nCmdShow)
{
    HWND                        hWnd;
    WNDCLASS                    wc;

    //Ustawienie wartoci komrek symulacji
    memset(Plytka1, 100, 240*240*sizeof(unsigned short));
    memset(Plytka2, 100, 240*240*sizeof(unsigned short));

    //Ustawienie parametrw WindowClass
    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = NULL;
    wc.hCursor = NULL;
    wc.hbrBackground = (HBRUSH )GetStockObject(BLACK_BRUSH);
    wc.lpszClassName = L"Chaos";
	wc.lpszMenuName = NULL;
    RegisterClass(&wc);
	
    //Tu przydaaby si obsuga bdw :-)

    //Tworzymy okno penoekranowe
    hWnd = CreateWindowEx(WS_EX_TOPMOST,
                          L"Chaos",
                          TITLE,
                          WS_POPUP,
                          0,
                          0,
                          GetSystemMetrics(SM_CXSCREEN),
                          GetSystemMetrics(SM_CYSCREEN),
                          NULL,
                          NULL,
                          hInstance,
                          NULL);
    if (!hWnd)
    {
	//Tu z kolei mona by wywietli jaki komunikat o bdzie :-)
	return FALSE;
    }
        
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);
    SetFocus(hWnd);

	//Wczamy GAPI
	if (GXOpenDisplay( hWnd, GX_FULLSCREEN) == 0) //Moe te by GXOpenDisplay( NULL, GX_FULLSCREEN), ale to nie zadziaa z emulatorem GAPI
	return FALSE;

	//Tu badamy, jakie s parametry wywietlacza
	gx_displayprop = GXGetDisplayProperties();

	//Sprawdzamy gbi kolorw, dla wartoci innych ni 16 bitowa adresowanie
	//pamici ekranu staje si nieco bardziej skomplikowane...
	if (gx_displayprop.cBPP != 16)
	{
		GXCloseDisplay();
		MessageBox(hWnd,L"Niestety, to nie jest wywietlacz 16 bitowy",L"Bd!", MB_OK);
		return FALSE;
	}


	//Przygotowujemy palet kolorw

	//R 0 - 32
	//G 0 - 64
	//B 0 - 32

	int r,g,b;
	for (int i=0;i<=255;i++) 
        {
		r = -4*abs(-1*(i-170))+255; if (r<0) r=0;        
        g = abs(-4*abs(-1*(i-127))+255); if (g<0) g=0;
        b = -4*abs(-1*(i-20))+255; if (b<0) b=0;
		
		Palette[i] = ComputeColor(r/8,g/4,b/8);
	}

	return TRUE;
}//of InitApp

//Program entry point
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nShowCmd )
{
    MSG msg;

    if (InitApp(hInstance, nShowCmd) != TRUE) return FALSE;	                

	while( 1 )
        {

        if( PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) )
        {
            if( !GetMessage( &msg, NULL, 0, 0 ) )
            {
                return msg.wParam;
            }
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            
        }
        else
        {
            UpdateFrame(); //Kolejny krok symulacji
        }        
    }
    
}//of WinMain